import numpy as np, pandas as pd
from . import cpp_funs
from scipy.sparse import issparse, csr_array
from warnings import warn
import re
import multiprocessing
import ctypes

__all__ = ["calc_reco_metrics", "split_reco_train_test"]

def _as_row_major(X):
    if X.flags["C_CONTIGUOUS"]:
        return X, X.shape[1]
    if X.strides[1] != X.dtype.itemsize:
        return np.ascontiguousarray(X, dtype=X.dtype), X.shape[1]
    return X, int(X.strides[0] / X.itemsize)

def _cast_to_dtype(X, use_float):
    if not use_float:
        if X.dtype != ctypes.c_double:
            X = X.astype(ctypes.c_double)
    else:
        if X.dtype != ctypes.c_float:
            X = X.astype(ctypes.c_float)
    return X

def _cast_indices_to_int32(X):
    if (X.indptr.dtype != np.int32) or (X.indices.dtype != np.int32):
        X = X.copy()
        X.indptr = X.indptr.astype(np.int32)
        X.indices = X.indices.astype(np.int32)
    return X

def _as_csr(X):
    if issparse(X):
        if X.format != "csr":
            X = X.tocsr()
        X.sort_indices()
        return X
    else:
        return csr_array(X)


def calc_reco_metrics(
    X_train, X_test,
    A, B,
    k = 5,
    item_biases = None,
    as_df = True,
    precision = True,
    trunc_precision = False,
    recall = False,
    average_precision = True,
    trunc_average_precision = False,
    ndcg = True,
    hit = False,
    rr = False,
    roc_auc = False,
    pr_auc = False,
    all_metrics = False,
    rename_k = True,
    break_ties_with_noise = True,
    min_pos_test = 1,
    min_items_pool = 2,
    consider_cold_start = True,
    cumulative = False,
    nthreads = -1,
    seed = 1
):
    """
    Calculate Recommendation Quality Metrics

    Calculates recommendation quality metrics for implicit-feedback
    recommender systems (fit to user-item interactions data such as "number of
    times that a user played each song in a music service") that are based on
    low-rank matrix factorization or for which predicted scores can be reduced to
    a dot product between user and item factors/components.
    
    These metrics are calculated on a per-user basis, by producing a ranking of the
    items according to model predictions (in descending order), ignoring the items
    that are in the training data for each user. The items that were not consumed
    by the user (not present in ``X_train`` and not present in ``X_test``) are considered
    "negative" entries, while the items in ``X_test`` are considered "positive" entries,
    and the items present in ``X_train`` are ignored for these calculations.
    
    The metrics that can be calculated by this function are:

    ``P@K`` ("precision-at-k")
        Denotes the proportion of items among the top-K
        recommended (after excluding those that were already in the training data)
        that can be found in the test set for that user:

        :math:`P@K = \\frac{1}{k} \\sum_{i=1}^k r_i \\in \\mathcal{T}`
    
        This is perhaps the most intuitive and straightforward metric, but it can
        present a lot of variation between users and does not take into account
        aspects such as the number of available test items or the specific ranks at
        which they are shown.

    ``TP@K`` (truncated precision-at-k)
        A truncated or standardized version
        of the precision metric, which will divide instead by the minimum between
        ``k`` and the number of test items:
        
        :math:`TP@K = \\frac{1}{\\min\\{k, \\mathcal{T}\\}} \\sum_{i=1}^k r_i \\in \\mathcal{T}`
        
        **Note:** many papers and libraries call this the "P@K" instead. The
        "truncated" prefix is a non-standard nomenclature introduced here to
        differentiate it from the P@K metric that is calculated by this and
        other libraries.

    ``R@K`` ("recall-at-k")
        Proportion of the test items that are retrieved
        in the top-K recommended list. Calculation is the same as precision, but the
        division is by the number of test items instead of ``k``:
        
        :math:`R@K = \\frac{1}{|\\mathcal{T}|} \\sum_{i=1}^k r_i \\in \\mathcal{T}`

    ``AP@K`` ("average precision-at-k")
        Precision and recall look at all the items
        in the top-K equally, whereas one might want to take into account also the ranking
        within this top-K list, for which this metric comes in handy.
        "Average Precision" tries to reflect the precisions that would be obtained at
        different recalls:
        
        :math:`AP@K = \\frac{1}{|\\mathcal{T}|} \\sum_{i=1}^k (r_i \\in \\mathcal{T}) \\cdot P@i`
        
        This is a metric which to some degree considers precision, recall, and rank within
        top-K. Intuitively, it tries to approximate the are under a precision-recall
        tradeoff curve.
        
        The average of this metric across users is known as "Mean Average Precision"
        or "MAP@K".
        
        **IMPORTANT:** many authors define AP@K differently, such as dividing by
        the minimum between ``k`` and the number of test items instead, or as the average
        for P@1..P@K (either as-is or stopping after already retrieving all the test
        items) - here, the second version is offered as different metric instead.
        This metric is likely to be a source of mismatches when comparing against
        other libraries due to all the different defintions used by different authors.

    ``TAP@K`` (truncated average precision-at-k)
        A truncated version of the
        AP@K metric, which will instead divide it by the minimum between ``k`` and the
        number of test items.
        
        Many other papers and libraries call this the "average precision" instead.

    ``NDCG@K`` (normalized discounted cumulative gain at K)
        A ranking metric
        calculated by first determining the following:
        
        :math:`\\sum_{i=1}^k \\frac{C_i}{log_2(i+1)}`
        
        Where :math:`C_i` denotes the confidence score for an item (taken as the value
        in ``X_test`` for that item), with :math:`i` being the item ranked at a given position
        for a given user according to the model. This metric is then standardized by
        dividing by the maximum achievable such score given the test data.
        
        Unlike the other metrics:

            - It looks not only at the presence or absence of items, but also at their
              confidence score.

            - It can handle data which contains "dislikes" in the form of negative
              values (see caveats below).
        
        If there are only non-negative values in ``X_test``, this metric will be bounded
        between zero and one.

        A note about negative values: the NDCG metric assumes that all the values are
        non-negative. This implementation however can accommodate situations in which
        a very small fraction of the items have negative values, in which case:
        (a) it will standardize the metric by dividing by a number which does not
        consider the negative values in its sum; (b) it will be set to ``nan`` if there
        are no positive values. Be aware however that NDCG loses some of its desirable
        properties in the presence of negative values.

    ``Hit@K`` ("hit-at-k")
        Indicates whether any of the top-K recommended items
        can be found in the test set for that user. The average across users is typically
        referred to as the "Hit Rate".
        
        This is a binary metric (it is either zero or one, but can also be ``nan`` when
        it is not possible to calculate, just like the other metrics).

    ``RR@K`` ("reciprocal-rank-at-k")
        Inverse rank (one divided by the rank)
        of the first item among the top-K recommended that is in the test data.
        The average across users is typically referred to as the "Mean Reciprocal Rank"
        or MRR.

        If none of the top-K recommended items are in the test data, will be set to zero.

    ``ROC-AUC`` (area under the receiver-operating characteristic curve)
        See the
        `Wikipedia entry <https://en.wikipedia.org/wiki/Receiver_operating_characteristic>`_
        for details. This metric considers the full ranking of items
        rather than just the top-K. It is bounded between zero and one, with a value of
        0.5 corresponding to a random order and a value of 1 corresponding to a perfect
        ordering (i.e. every single positive item has a higher predicted score than every
        single negative item).
        
        Be aware that there are different ways of calculating AUC, with some methods
        having higher precision than others. This implementation uses a fast formula
        which implies dividing two large numbers, and as such might not be as precise
        to the same number of decimals as the trapezoidal method used by e.g. scikit-learn.

    ``PR-AUC`` (area under the precision-recall curve)
        While ROC AUC provides an
        overview of the overall ranking, one is typically only interested in how well it
        retrieves test items within top ranks, and for this the area under the
        precision-recall curve can do a better job at judging rankings, albeit the metric
        itself is not standardized, and under the worst possible ranking, it does not
        evaluate to zero.
        
        The metric is calculated using the fast but not-so-precise rectangular method,
        whose formula corresponds to the AP@K metric with K=N. Some papers and libraries
        call this the average of this metric the "MAP" or "Mean Average Precision" instead
        (without the "@K").

    Metrics can be calculated for a given value of ``k`` (e.g. "P@3"), or for
    values ranging from 1 to ``k`` (e.g. ["P@1", "P@2", "P@3"]).

    This package does **NOT** cover other more specialized metrics. One might
    also want to look at other indicators such as:
        - Metrics that look at the rareness of the items recommended
          (to evaluate so-called "serendipity").
        - Metrics that look at "discoverability".
        - Metrics that take into account the diversity of the ranked lists.
    

    Metrics for a given user will be set to ``nan`` in the following
    situations:
        - All the rankeable items have the exact same predicted score.
        - One or more of the predicted scores evaluates to ``nan``.
        - There are only negative entries (no non-zero entries in the test data).
        - The number of available items to rank (between positive and negative) is
          smaller than the requested ``k``, and the metric is not affected by the exact order
          within the top-K items (i.e. precision, recall, hit, will be ``nan`` if there's
          ``k`` or fewer items after discarding those from the training data).
        - There are inconsistencies in the data (e.g. number of entries being greater
          than the number of columns in "X", meaning the matrices do not constitute valid CSR).
        - A user does not meet the minimum criteria set by the configurable parameters
          for this function.
        - There are only positive entries (i.e. the user already consumed all the items).
          In this case, "NDCG@K" will still be calculated, while the rest will be set
          to ``nan``.
    
    The NDCG@K metric with ``cumulative=True`` will have lower decimal precision
    than with ``cumulative=False`` when using ``np.float32`` inputs - this is extremely
    unlikely to be noticeable in typical datasets and small ``k``, but for large ``k``
    and large (absolute) values in ``X_test``, it might make a difference after a
    couple of decimal points.

    Parameters
    ----------
    X_train : CSR(m, n) or None
        Training data for user-item interactions, with users denoting rows,
        items denoting columns, and values corresponding to confidence scores.
        Entries in ``X_train`` and ``X_test`` for each user should not intersect (that is,
        if an item is the training data as a non-missing entry, it should not be in
        the test data as non-missing, and vice versa).
        
        Should be passed as a sparse array/matrix in CSR format
        from SciPy. Items not consumed by the user should not
        be present in this matrix.
        
        Alternatively, if there is no training data, can pass ``None``, in which case it
        will look only at the test data.
        
        This matrix and ``X_test`` are not meant to contain negative values, and if
        ``X_test`` does contain any, it will still be assumed for all metrics other than
        NDCG that such items are deemed better for the user than the missing/zero-valued
        items (that is, implicit feedback is not meant to signal dislikes).
    X_test : CSR(m, n)
        Test data for user-item interactions. Same format as ``X_train``.
    A : array(m, p) or None
        The user factors, as a NumPy array. If the number of users is 'm' and the number of
        factors is 'p', should have dimension '(m, p)'.
        If both ``A`` and ``B`` are passed with ``np.float32`` dtype, will do the
        calculations in single precision (which is faster and uses less memory) and
        output the  calculated metrics in that same precision, otherwise will do them
        in ``np.float64`` precision.
        
        It is assumed that the model score for a given item ``j`` for user ``i`` is
        calculated as the inner product or dot product between the corresponding vectors
        :math:`\\mathbf{a}_i \\cdot \\mathbf{b}_j`
        (rows ``i`` and ``j`` of ``A`` and ``B``, respectively),
        with higher scores meaning that the item is deemed better for
        that user, and the top-K recommendations produced by ranking these scores in
        descending order.
        
        Alternatively, for evaluation of non-personalized models, can pass ``None`` here
        and for ``B``, in which case ``item_biases`` must be passed.
    B : array(n, p) or None
        The item factors, in the same format as ``A``.
    k : int
        The number of top recommendations to consider for the metrics (as
        in "precision-at-k" or "P@K").
    item_biases : None or array(n,)
        Optional item biases/intercepts (fixed base score that is
        added to the predictions of each item). If present, it will append them to ``B``
        as an extra factor while adding a factor of all-ones to ``A``.

        Alternatively, for non-personalized models which have only item-by-item scores,
        can pass ``None`` for ``A`` and ``B`` while passing only ``item_biases``.
    as_df : bool
        Whether to output the result as a pandas DataFrame. If passing ``False``,
        the results will be returned as a dict of NumPy arrays - either 1d or 2d depending on
        what is passed for ``cumulative``.
    precision : bool
        Whether to calculate precision metrics or not.
    trunc_precision : bool
        Whether to calculate truncated precision metrics or not.
        Note that this is output as a separate metric from "precision" and they are not
        mutually exclusive options.
    recall : bool
        Whether to calculate recall metrics or not.
    average_precision : bool
        Whether to calculate average precision metrics or not.
    trunc_average_precision : bool
        Whether to calculate truncated average
        precision metrics or not. Note that this is output as a separate metric from
        "average_precision" and they are not mutually exclusive options.
    ndcg : bool
        Whether to calculate NDCG (normalized discounted cumulative gain)
        metrics or not.
    hit : bool
        Whether to calculate Hit metrics or not.
    rr : bool
        Whether to calculate RR (reciprocal rank) metrics or not.
    roc_auc : bool
        Whether to calculate ROC-AUC (area under the ROC curve) metrics or not.
    pr_auc : bool
        Whether to calculate PR-AUC (area under the PR curve) metrics or not.
    all_metrics : bool
        Passing ``True`` here is equivalent to passing ``True`` to all the
        calculable metrics.
    rename_k : bool
        If passing ``as_df=True`` and ``cumulative=False``, whether to rename
        the 'k' in the resulting column names to the actual value of 'k' that was used
        (e.g. "P@K" -> "P@5").
    break_ties_with_noise : bool
        Whether to add a small amount of noise
        :math:`\\sim \\text{Uniform}(-10^{-12}, 10^{-12})` in order to break ties
        at random, in case there are any ties in the ranking. This is not recommended unless
        one expects ties (can happen if e.g. some factors are set to all-zeros for some items),
        as it has the potential to slightly alter the ranking.
    min_pos_test : int
        Minimum number of positive entries
        (non-zero entries in the test set) that users need to have in
        order to calculate metrics for that user.
        If a given user does not meet the threshold, the metrics
        will be set to ``nan``.
    min_items_pool : int
        Minimum number of items (sum of positive and negative items)
        that a user must have in order to
        calculate metrics for that user. If a given user does not meet the threshold,
        the metrics will be set to ``nan``.
    consider_cold_start : bool
        Whether to calculate metrics in situations in
        which some user has test data but no positive
        (non-zero) entries in the training data. If passing ``False`` and such cases are
        encountered, the metrics will be set to ``nan``.

        Will be automatically set to ``True`` when passing ``None`` for ``X_train``.
    cumulative : bool
        Whether to calculate the metrics cumulatively
        (e.g. [P@1, P@2, P@3] if passing ``k=3``)
        for all values up to ``k``, or only for a single desired ``k``
        (e.g. only P@3 if passing ``k=3``).
    nthreads : int
        Number of parallel threads to use.
        Parallelization is done at the user level, so passing
        more threads than there are users will not result in a speed up. Be aware that, the more
        threads that are used, the higher the memory consumption.

        If passing a negative integer, the total number of threads to use will be determined
        in the same way as joblib:
            nthreads_use = available_threads + 1 + nthreads
    seed : int, RandomState, or Generator
        Seed used for random number generation. Only used when passing
        ``break_ties_with_noise=True``.

        Can be passed as an integer, or as a NumPy ``RandomState`` or
        ``Generator`` object, which will be used to draw a single random
        integer.

    Returns
    -------
    metrics : DataFrame or dict
        Will return the calculated metrics on a per-user basis (each user
        corresponding to a row):
            - If passing ``as_df=True`` (the default), the result will be a DataFrame,
              with the columns named according to the metric they represent (e.g. "P@3",
              see below for the other names that they can take). Depending on the value
              passed for ``rename_k``, the column names might end in "K" or in the number
              that was passed for ``k`` (e.g "P@3" or "P@K").

              If passing ``cumulative=True``, they will have names ranging from 1 to ``k``.
            - If passing ``as_df=False``, the result will be a list with entries named
              according to each metric, with ``k`` as letter rather than number (``P@K``,
              ``TP@K``, ``R@K``, ``AP@K``, ``TAP@K``, ``NDCG@K``, ``Hit@K``, ``RR@K``,
              ``ROC_AUC``, ``PR_AUC``), plus an additional entry with the actual ``k``.

              The values under each entry will be 1d arrays if passing
              ``cumulative=False``, or 2d arrays (users corresponding to rows) if passing
              ``cumulative=True``.
        
        The "ROC-AUC" and "PR-AUC" metrics will be named just "ROC_AUC" and "PR_AUC",
        since they are calculated for the full ranked predictions without stopping at ``k``.
    """
    if all_metrics:
        precision = True
        trunc_precision = True
        recall = True
        average_precision = True
        trunc_average_precision = True
        ndcg = True
        hit = True
        rr = True
        roc_auc = True
        pr_auc = True

    if (item_biases is not None) and (isinstance(item_biases, pd.Series)):
        item_biases = item_biases.to_numpy()

    if (A is None) != (B is None):
        raise ValueError("'A' and 'B' must either be passed together or passed as 'None' together.")
    if (A is None):
        if (item_biases is None):
            raise ValueError("Must pass item biases if not passing factors.")
        A = np.ones((X_test.shape[0],1), dtype=ctypes.c_double, order="C")
        B = np.ascontiguousarray(item_biases, dtype=ctypes.c_double).reshape((-1,1))
        item_biases = None

    assert isinstance(A, np.ndarray)
    assert isinstance(B, np.ndarray)
    assert issparse(X_test)

    if X_test.shape[0] >= np.iinfo(np.int32).max:
        raise ValueError("Number of test user is larger than maximum supported.")
    if X_test.shape[1] >= np.iinfo(np.int32).max:
        raise ValueError("Number of items is larger than maximum supported.")
    if not X_test.data.shape[0]:
        raise ValueError("'X_test' is empty.")

    if len(A.shape) != 2:
        raise ValueError("'A' must be a 2-dimensional array.")
    if len(B.shape) != 2:
        raise ValueError("'B' must be a 2-dimensional array.")
    if A.shape[1] != B.shape[1]:
        raise ValueError("'A' and 'B' must have the same number of columns.")
    if (not A.shape[0]) or (not A.shape[1]) or (not B.shape[1]) or (not X_test.shape[0]) or (not X_test.shape[1]):
        raise ValueError("Input matrices cannot be empty.")
    if A.shape[0] < X_test.shape[0]:
        raise ValueError("Number of users in 'A' and 'X_test' does not match.")
    if B.shape[0] < X_test.shape[1]:
        raise ValueError("Number of items in 'B' and 'X_test' does not match.")

    if (A.shape[0] > X_test.shape[0]):
        warn("'A' has more users than 'X_test'.")
        A = A[:X_test.shape[0], :]
    if (B.shape[0] > X_test.shape[1]):
        warn("'B' has more items than 'X_test'.")
        B = B[:X_test.shape[1], :]

    use_float = (A.dtype == ctypes.c_float) and (B.dtype == ctypes.c_float)

    if X_train is None:
        X_train = csr_array(X_test.shape, dtype=ctypes.c_double if not use_float else ctypes.c_float)
        consider_cold_start = True

    assert issparse(X_train)
    assert X_train.shape[1] == X_test.shape[1]
    if (X_train.shape[0] < X_test.shape[0]):
        raise ValueError("'X_train' and 'X_test' should have the same number of rows.")
    elif (X_train.shape[0] > X_test.shape[0]):
        warn("'X_train' mas more rows than 'X_test'.")

    as_df = bool(as_df)
    break_ties_with_noise = bool(break_ties_with_noise)
    consider_cold_start = bool(consider_cold_start)
    cumulative = bool(cumulative)
    rename_k = bool(rename_k)
    
    precision = bool(precision)
    trunc_precision = bool(trunc_precision)
    recall = bool(recall)
    average_precision = bool(average_precision)
    trunc_average_precision = bool(trunc_average_precision)
    ndcg = bool(ndcg)
    hit = bool(hit)
    rr = bool(rr)
    roc_auc = bool(roc_auc)
    pr_auc = bool(pr_auc)
    
    if (not precision) and (not average_precision) and (not ndcg) and (not hit) and (not rr) and (not roc_auc):
        raise ValueError("Must pass at least one metric to calculate.")

    if isinstance(seed, np.random.RandomState):
        seed = int(seed.randint(np.iinfo(np.int32).max))
    elif isinstance(seed, np.random.Generator):
        seed = int(seed.integers(np.iinfo(np.int32).max))

    nthreads = int(nthreads)
    seed = int(seed)
    k = int(k)
    min_pos_test = int(min_pos_test)
    min_items_pool = int(min_items_pool)

    assert seed >= 1
    assert k >= 1
    assert min_pos_test >= 1
    assert min_items_pool >= 1

    if nthreads is None:
        nthreads = 1
    elif nthreads < 0:
        nthreads = multiprocessing.cpu_count() + 1 + nthreads
    assert nthreads > 0

    if (nthreads > 1) and (not cpp_funs._get_has_openmp()):
        msg_omp  = "Attempting to use more than 1 thread, but "
        msg_omp += "package was built without multi-threading "
        msg_omp += "support - see the project's GitHub page for "
        msg_omp += "more information."
        warn(msg_omp)

    if k > X_test.shape[1]:
        raise ValueError("'k' should be smaller than the number of items.")

    if item_biases is not None:
        assert isinstance(item_biases, np.ndarray)
        if (len(item_biases.shape) > 2):
            raise ValueError("'item_biases' should be a 1-d array.")
        if (len(item_biases.shape) != 1):
            item_biases = item_biases.reshape(-1)
        if (not item_biases.shape[0]):
            raise ValueError("'item_biases' is empty.")
        if item_biases.shape[0] < X_test.shape[1]:
            raise ValueError("Number of items in 'item_biases' must match with 'X_test'.")
        if item_biases.shape[0] > X_test.shape[1]:
            item_biases = item_biases[:X_test.shape[1]]
            warn("'item_biases' has more items than 'X_test'.")
        
        item_biases = _cast_to_dtype(item_biases, use_float)
        A = np.c_[A, np.ones((A.shape[0], 1), dtype=ctypes.c_double if not use_float else ctypes.c_float)]
        B = np.c_[B, item_biases.reshape((-1,1))]
        item_biases = None

    X_train = _as_csr(X_train)
    X_test = _as_csr(X_test)
    
    X_train = _cast_indices_to_int32(X_train)
    X_test = _cast_to_dtype(X_test, use_float)
    X_test = _cast_indices_to_int32(X_test)
    A = _cast_to_dtype(A, use_float)
    A, lda = _as_row_major(A)
    B = _cast_to_dtype(B, use_float)
    B, ldb = _as_row_major(B)

    p_at_k, tp_at_k, r_at_k, ap_at_k, tap_at_k, ndcg_at_k, \
    hit_at_k, rr_at_k, roc_auc, pr_auc = \
        cpp_funs.calc_reco_metrics(
            A, lda,
            B, ldb,
            X_train, X_test,
            k_metrics = k,
            precision = precision,
            trunc_precision = trunc_precision,
            recall = recall,
            average_precision = average_precision,
            trunc_average_precision = trunc_average_precision,
            ndcg = ndcg,
            hit = hit,
            rr = rr,
            roc_auc = roc_auc,
            pr_auc = pr_auc,
            break_ties_with_noise = break_ties_with_noise,
            min_pos_test = min_pos_test,
            min_items_pool = min_items_pool,
            consider_cold_start = consider_cold_start,
            cumulative = cumulative,
            nthreads = nthreads,
            seed = seed
        )

    out = dict()
    if p_at_k.shape[0]:
        out["P@K"] = p_at_k
    if tp_at_k.shape[0]:
        out["TP@K"] = tp_at_k
    if r_at_k.shape[0]:
        out["R@K"] = r_at_k
    if ap_at_k.shape[0]:
        out["AP@K"] = ap_at_k
    if tap_at_k.shape[0]:
        out["TAP@K"] = tap_at_k
    if ndcg_at_k.shape[0]:
        out["NDCG@K"] = ndcg_at_k
    if hit_at_k.shape[0]:
        out["Hit@K"] = hit_at_k
    if rr_at_k.shape[0]:
        out["RR@K"] = rr_at_k
    if roc_auc.shape[0]:
        out["ROC_AUC"] = roc_auc
    if pr_auc.shape[0]:
        out["PR_AUC"] = pr_auc

    if not as_df:
        out["K"] = k

    if as_df:
        if not cumulative:
            out = pd.DataFrame(out)
            if rename_k:
                out.columns = out.columns.str.replace("@K$", "@"+str(k), regex=True)
        else:
            out = [
                pd.DataFrame(v, columns=[re.sub("(@)K$", r"\1", k)+(str(i+1) if bool(re.search("@K$", k)) else "")
                                         for i in range(v.shape[1])])
                for k,v in out.items()
            ]
            out = pd.concat(out, axis=1)

    return out

def split_reco_train_test(
    X,
    split_type = "separated",
    users_test_fraction = 0.1,
    max_test_users = 10000,
    items_test_fraction = 0.3,
    min_items_pool = 2,
    min_pos_test = 1,
    consider_cold_start = False,
    seed = 1
):
    """
    Create Train-Test Splits of Implicit-Feedback Data

    Creates train-test splits of implicit-feedback data
    (recorded user-item interactions) for fitting and evaluating models for
    recommender systems.

    These splits choose "test users" and "items for a given user" separately,
    offering three modes of splitting the data:

        * Creating training and testing sets for each user in the data (when
          passing ``split_type='all'``).

          This is meant for cases in which the number of users is small or the users to
          test have already been selected (e.g. one typically does not want to create
          a train-test split which would leave one item for the user in the training data
          and zero in the test set, or would want to have other minimum criteria for the
          test set to be usable). Typically, one would want to take only a sub-sample
          of users for evaluation purposes, as calculating recommendation quality metrics
          can take a very long time.

        * Selecting a sub-set of users for testing, for which training and testing
          data splits will be generated, while leaving the remainder of users with all
          the data for model fitting (when passing ``split_type='separated'``).

          This is meant to be used for fitting a model to the remainder
          of the data, then generating latent factors (assuming a low-rank matrix factorization
          model) or top-K recommendations for the test users given their training data,
          and evaluating these recommendations on the test data for each user (which can be
          done through the function ``calc_reco_metrics``).

        * Selecting a sub-set of users for testing as above, but adding those users to
          the training data, in which case they will be the first rows (when passing
          ``split_type='joined'``).

          This is meant to be used for fitting a model to all such training data, and
          then evaluating the produced user factors or top-K recommendations for the test
          users against the test data.

          It is recommended to use the "separated" mode, as it
          is more reflective of real scenarios, but some models or libraries do not have the
          capabilities for producing factors/recommendations for users which where not in the
          training data, and this option then comes in handy.

    Parameters
    ----------
    X : CSR(m, n)
        The implicit feedback data to split into training-testing-remainder
        for evaluating recommender systems. Should be passed as a sparse CSR array/matrix from
        SciPy, or will be converted to CSR if it isn't. Users should correspond to rows, items
        to columns, and non-zero values to observed user-item interactions.

        Note that the indices of this matrix will be sorted in-place if it is passed as CSR.
        The data must be of type ``np.float32`` or ``np.float64`` - will be converted to
        ``np.float64`` if it is of a different type.
    split_type : str
        Type of data split to generate. Allowed values are:
        "all", "separated", "joined" (see the function description above for more details).
    users_test_fraction : float or None
        Target fraction of the users to set as test (see the
        function documentation for more details). If the number represented by this fraction
        exceeds the number set by ``max_test_users``, then the actual number will be set to
        ``max_test_users``. Note however that the end result might end up containing fewer
        users if there are not enough users in the data meeting the minimum desired
        criteria.

        If passing ``None``, will not take a fraction, but will instead take the number that
        is passed for ``max_test_users``.

        Ignored when passing ``split_type='all'``.
    max_test_users : float or None
        Maximum number of users to set as test. Note that this will
        only be applied for choosing the minimum between this and
        'ncol(X)*users_test_fraction', while the actual number might end up being
        lower if there are not enough users meeting the desired minimum conditions.

        If passing ``None`` for ``users_test_fraction``, will interpret this as the number
        of test users to take.

        Ignored when passing ``split_type='all'``.
    items_test_fraction : float
        Target fraction of the data (items) to set for test
        for each user. Should be a number between zero and one (non-inclusive).
        The actual number of test entries for each user will be determined as
        "round(n_entries_user*items_test_fraction)", thus in a long-tailed distribution
        (typical for recommender systems), the actual fraction that will be obtained is
        likely to be slightly lower than what is passed here.

        Note that items are sampled independently for each user, thus the items that are
        in the test set for some users might be in the training set for different users.
    min_items_pool : int
        Minimum number of items (sum of positive and negative items)
        that a user must have in order to be eligible as test user.
    min_pos_test : int
        Minimum number of positive entries (non-zero entries in
        the test set) that users would need to have in order to be eligible as test user.
    consider_cold_start : bool
        Whether to still set users as eligible for test in
        situations in which some user would have test data but no positive (non-zero)
        entries in the training data. This will only happen when passing
        ``test_fraction>=0.5``.
    seed : int
        Seed to use for random number generation.

    Returns
    -------
    split : tuple
        Will return a tuple with two to four elements depending on the requested split
        type:
            * If passing ``split_type='all'``, will output 2 elements: "X_train" and "X_test",
              both of which will be sparse CSR array/matrices
              with the same number of rows and columns as the ``X`` that was passed as input.
            * If passing ``split_type='separated'``, will output 4 elements: "X_train" and "X_test"
              as above (but with a number of rows corresponding to the number of selected test
              users instead), plus another CSR array/matrix "X_rem" which will contain the
              data for the remainder of the users (those which were not selected for testing and
              on which the recommendation model is meant to be fitted), and finally "users_test"
              which will be an integer vector containing the indices of the users/rows in ``X``
              which were selected for testing. The selected test users will be in sorted order,
              and the remaining data will remain in sorted order with the test users removed
              (e.g. if there's 5 users, with the second and fifth selected for testing, then
              "X_train" and "X_test" will contain rows [2,5] of ``X``, while "X_rem" will contain
              rows [1,3,4]).
            * If passing ``split_type='joined'``, will output 3 elements: same as above but will
              not contain "X_rem" - instead, "X_train" will be the concatenation of
              "X_train" and "X_rem",
              with "X_train" coming first (e.g. if there's 5 users, with the second and fifth
              selected for testing, then "X_test" will contain rows [2,5] of "X", while
              "X_train" will contain rows [2,5,1,3,4], in that order).
    """
    if (max_test_users is None) or (max_test_users == 0):
        max_test_users = X.shape[0]
    
    assert max_test_users > 0
    assert seed >= 0
    assert min_pos_test >= 0
    assert min_items_pool >= 0
    
    max_test_users = int(max_test_users)
    seed = int(seed)
    min_pos_test = int(min_pos_test)
    min_items_pool = int(min_items_pool)
    
    if users_test_fraction is not None:
        assert (users_test_fraction > 0) and (users_test_fraction < 1)
        users_test_fraction = float(users_test_fraction)

    assert (items_test_fraction > 0) and (items_test_fraction < 1)
    items_test_fraction = float(items_test_fraction)

    assert split_type in ("all", "separated", "joined")

    consider_cold_start = bool(consider_cold_start)


    if min_pos_test >= X.shape[1]:
        raise ValueError("'min_pos_test' must be smaller than the number of columns in 'X'.")

    if min_items_pool >= X.shape[1]:
        raise ValueError("'min_items_pool' must be smaller than the number of columns in 'X'.")

    if split_type != "all":
        if X.shape[0] < 2:
            raise ValueError("'X' has less than 2 rows.")
        if users_test_fraction is not None:
            n_users_take = X.shape[0] * users_test_fraction
            if n_users_take < 1:
                warn("Desired fraction of test users implies <1, will select 1 user.")
                n_users_take = 1
            n_users_take = round(n_users_take)
            n_users_take = min(n_users_take, max_test_users)
        else:
            if max_test_users > X.shape[0]:
                warn("'max_test_users' is larger than number of users. Will take all.")
            n_users_take = min(max_test_users, X.shape[0])

    X = _as_csr(X)
    if (not X.shape[0]) or (not X.shape[1]):
        raise ValueError("'X' cannot be empty.")
    if X.dtype not in (np.float32, np.float64):
        X = X.astype(np.float64)
    if not X.data.shape[0]:
        raise ValueError("'X' contains no non-zero entries.")
    X = _cast_indices_to_int32(X)

    if split_type == "all":
        return cpp_funs.split_csr_selected_users(X, items_test_fraction, seed)
    elif split_type == "separated":
        return cpp_funs.split_csr_separated_users(
            X,
            n_users_take,
            items_test_fraction,
            consider_cold_start,
            min_items_pool,
            min_pos_test,
            True,
            seed
        )
    elif split_type == "joined":
        return cpp_funs.split_csr_separated_users(
            X,
            n_users_take,
            items_test_fraction,
            consider_cold_start,
            min_items_pool,
            min_pos_test,
            False,
            seed
        )
    else:
        raise ValueError("Unexpected error.")



import setuptools
from setuptools import setup, Extension
import numpy as np
import sys, os, subprocess, warnings, re
from Cython.Distutils import build_ext

found_omp = True
def set_omp_false():
    global found_omp
    found_omp = False

## Modify this to make the output of the compilation tests more verbose
silent_tests = not (("verbose" in sys.argv)
                    or ("-verbose" in sys.argv)
                    or ("--verbose" in sys.argv))

## Workaround for python<=3.9 on windows
try:
    EXIT_SUCCESS = os.EX_OK
except AttributeError:
    EXIT_SUCCESS = 0

class build_ext_subclass( build_ext ):
    def build_extensions(self):
        is_msvc = self.compiler.compiler_type == "msvc"
        is_clang = hasattr(self.compiler, 'compiler_cxx') and ("clang++" in self.compiler.compiler_cxx)
        is_windows = sys.platform[:3].lower() == "win"

        compiler = self.compiler.compiler_type
        if is_msvc:
            for e in self.extensions:
                e.extra_compile_args += ['/O2', '/openmp', '/GL', '/std:c++14', '/fp:contract', '/fp:except-']
        else:
            if not self.check_for_variable_dont_set_march() and not self.check_cflags_contain_arch():
                self.add_march_native()
            else:
                if not is_windows:
                    self.add_link_time_optimization()
                for e in self.extensions:
                    e.define_macros += [("USE_BLAS", None)]
            self.add_openmp_linkage()
            self.add_restrict_qualifier()
            self.add_no_math_errno()
            self.add_no_trapping_math()
            self.add_O3()
            if not is_clang:
                self.add_std_cpp11()
            else:
                self.add_std_cpp17()

            # for e in self.extensions:
                # e.extra_compile_args += ['-O3', '-fopenmp', '-march=native', '-std=c++11']
                # e.extra_link_args += ['-fopenmp']

                # if is_clang:
                #     e.extra_compile_args += ['-std=c++17']

                # e.extra_compile_args += ['-O3', '-fopenmp', '-march=native', '-std=c++17']
                # e.extra_link_args += ['-fopenmp=libiomp5']

                # e.extra_compile_args += ['-fsanitize=address', '-static-libasan', '-ggdb', '-std=c++14']
                # e.extra_link_args += ['-fsanitize=address', '-static-libasan']

                # e.extra_compile_args += ['-O0', '-std=c++11', '-ggdb']
                # e.define_macros += [("DEBUG", None), ("_GLIBCXX_ASSERTIONS", None)]

            ## Note: apple will by default alias 'gcc' to 'clang', and will ship its own "special"
            ## 'clang' which has no OMP support and nowadays will purposefully fail to compile when passed
            ## '-fopenmp' flags. If you are using mac, and have an OMP-capable compiler,
            ## comment out the code below, or set 'use_omp' to 'True'.
        build_ext.build_extensions(self)

    def check_cflags_contain_arch(self):
        if ("CFLAGS" in os.environ) or ("CXXFLAGS" in os.environ):
            has_cflags = "CFLAGS" in os.environ
            has_cxxflags = "CXXFLAGS" in os.environ
            arch_list = [
                "-march", "-mcpu", "-mtune", "-msse", "-msse2", "-msse3",
                "-mssse3", "-msse4", "-msse4a", "-msse4.1", "-msse4.2",
                "-mavx", "-mavx2", "-mavx512"
            ]
            for flag in arch_list:
                if has_cflags and flag in os.environ["CFLAGS"]:
                    return True
                if has_cxxflags and flag in os.environ["CXXFLAGS"]:
                    return True
        return False

    def check_for_variable_dont_set_march(self):
        return "DONT_SET_MARCH" in os.environ

    def add_march_native(self):
        is_apple = sys.platform[:3].lower() == "dar"
        args_march_native = ["-march=native", "-mcpu=native"]
        for arg_march_native in args_march_native:
            if self.test_supports_compile_arg(arg_march_native, with_c_comp=is_apple):
                for e in self.extensions:
                    e.extra_compile_args.append(arg_march_native)
                break

    def add_link_time_optimization(self):
        args_lto = ["-flto=auto", "-flto"]
        for arg_lto in args_lto:
            if self.test_supports_compile_arg(arg_lto):
                for e in self.extensions:
                    e.extra_compile_args.append(arg_lto)
                    e.extra_link_args.append(arg_lto)
                break

    def add_no_math_errno(self):
        arg_fnme = "-fno-math-errno"
        if self.test_supports_compile_arg(arg_fnme):
            for e in self.extensions:
                e.extra_compile_args.append(arg_fnme)
                e.extra_link_args.append(arg_fnme)

    def add_no_trapping_math(self):
        arg_fntm = "-fno-trapping-math"
        if self.test_supports_compile_arg(arg_fntm):
            for e in self.extensions:
                e.extra_compile_args.append(arg_fntm)
                e.extra_link_args.append(arg_fntm)

    def add_O3(self):
        arg_O3 = "-O3"
        if self.test_supports_compile_arg(arg_O3):
            for e in self.extensions:
                e.extra_compile_args.append(arg_O3)
                e.extra_link_args.append(arg_O3)

    def add_std_cpp11(self):
        arg_std_cpp11 = "-std=c++11"
        if self.test_supports_compile_arg(arg_std_cpp11):
            for e in self.extensions:
                e.extra_compile_args.append(arg_std_cpp11)
                e.extra_link_args.append(arg_std_cpp11)

    def add_std_cpp17(self):
        arg_std_cpp17 = "-std=c++17"
        if self.test_supports_compile_arg(arg_std_cpp17):
            for e in self.extensions:
                e.extra_compile_args.append(arg_std_cpp17)
                e.extra_link_args.append(arg_std_cpp17)

    def add_openmp_linkage(self):
        arg_omp1 = "-fopenmp"
        arg_omp2 = "-fopenmp=libomp"
        args_omp3 = ["-fopenmp=libomp", "-lomp"]
        arg_omp4 = "-qopenmp"
        arg_omp5 = "-xopenmp"
        is_apple = sys.platform[:3].lower() == "dar"
        args_apple_omp = ["-Xclang", "-fopenmp", "-lomp"]
        args_apple_omp2 = ["-Xclang", "-fopenmp", "-L/usr/local/lib", "-lomp", "-I/usr/local/include"]
        has_brew_omp = False
        if is_apple:
            try:
                res_brew_pref = subprocess.run(["brew", "--prefix", "libomp"], capture_output=True)
                if res_brew_pref.returncode == EXIT_SUCCESS:
                    brew_omp_prefix = res_brew_pref.stdout.decode().strip()
                    args_apple_omp3 = ["-Xclang", "-fopenmp", f"-L{brew_omp_prefix}/lib", "-lomp", f"-I{brew_omp_prefix}/include"]
                    has_brew_omp = True
            except Exception as e:
                pass



        if self.test_supports_compile_arg(arg_omp1, with_omp=True):
            for e in self.extensions:
                e.extra_compile_args.append(arg_omp1)
                e.extra_link_args.append(arg_omp1)
        elif is_apple and self.test_supports_compile_arg(args_apple_omp, with_omp=True):
            for e in self.extensions:
                e.extra_compile_args += ["-Xclang", "-fopenmp"]
                e.extra_link_args += ["-lomp"]
        elif is_apple and self.test_supports_compile_arg(args_apple_omp2, with_omp=True):
            for e in self.extensions:
                e.extra_compile_args += ["-Xclang", "-fopenmp"]
                e.extra_link_args += ["-L/usr/local/lib", "-lomp"]
                e.include_dirs += ["/usr/local/include"]
        elif is_apple and has_brew_omp and self.test_supports_compile_arg(args_apple_omp3, with_omp=True):
            for e in self.extensions:
                e.extra_compile_args += ["-Xclang", "-fopenmp"]
                e.extra_link_args += [f"-L{brew_omp_prefix}/lib", "-lomp"]
                e.include_dirs += [f"{brew_omp_prefix}/include"]
        elif self.test_supports_compile_arg(arg_omp2, with_omp=True):
            for e in self.extensions:
                e.extra_compile_args += ["-fopenmp=libomp"]
                e.extra_link_args += ["-fopenmp"]
        elif self.test_supports_compile_arg(args_omp3, with_omp=True):
            for e in self.extensions:
                e.extra_compile_args += ["-fopenmp=libomp"]
                e.extra_link_args += ["-fopenmp", "-lomp"]
        elif self.test_supports_compile_arg(arg_omp4, with_omp=True):
            for e in self.extensions:
                e.extra_compile_args.append(arg_omp4)
                e.extra_link_args.append(arg_omp4)
        elif self.test_supports_compile_arg(arg_omp5, with_omp=True):
            for e in self.extensions:
                e.extra_compile_args.append(arg_omp5)
                e.extra_link_args.append(arg_omp5)
        else:
            set_omp_false()

    # Note: in apple systems, it somehow might end up triggering the arguments with
    # the C compiler instead of the CXX compiler. What's worse, sometimes this compiler
    # thinks it's building for aarch64 even when executed in amd64.
    def test_supports_compile_arg(self, comm, with_omp=False, with_c_comp=False):
        is_supported = False
        try:
            if not hasattr(self.compiler, "compiler_cxx"):
                return False
            if not isinstance(comm, list):
                comm = [comm]
            print("--- Checking compiler support for option '%s'" % " ".join(comm))
            fname = "recometrics_compiler_testing.cpp"
            with open(fname, "w") as ftest:
                ftest.write(u"int main(int argc, char**argv) {return 0;}\n")
            try:
                if not isinstance(self.compiler.compiler_cxx, list):
                    cmd = list(self.compiler.compiler_cxx)
                else:
                    cmd = self.compiler.compiler_cxx
            except Exception:
                cmd = self.compiler.compiler_cxx
            if with_c_comp:
                if not isinstance(self.compiler.compiler, list):
                    cmd0 = list(self.compiler.compiler)
                else:
                    cmd0 = self.compiler.compiler
            if with_omp:
                with open(fname, "w") as ftest:
                    ftest.write(u"#include <omp.h>\nint main(int argc, char**argv) {return 0;}\n")
            try:
                val = subprocess.run(cmd + comm + [fname], capture_output=silent_tests).returncode
                is_supported = (val == EXIT_SUCCESS)
                if is_supported and with_c_comp:
                    val = subprocess.run(cmd0 + comm + [fname], capture_output=silent_tests).returncode
                    is_supported = (val == EXIT_SUCCESS)
            except Exception:
                is_supported = False
        except Exception:
            pass
        try:
            os.remove(fname)
        except Exception:
            pass
        return is_supported

    def add_restrict_qualifier(self):
        supports_restrict = False
        try:
            if not hasattr(self.compiler, "compiler_cxx"):
                return None
            print("--- Checking compiler support for '__restrict' qualifier")
            fname = "recometrics_compiler_testing.cpp"
            with open(fname, "w") as ftest:
                ftest.write(u"int main(int argc, char**argv) {return 0;}\n")
            try:
                if not isinstance(self.compiler.compiler_cxx, list):
                    cmd = list(self.compiler.compiler_cxx)
                else:
                    cmd = self.compiler.compiler_cxx
            except Exception:
                cmd = self.compiler.compiler_cxx
            try:
                with open(fname, "w") as ftest:
                    ftest.write(u"int main(int argc, char**argv) {double *__restrict x = 0; return 0;}\n")
                val = subprocess.run(cmd + comm + [fname], capture_output=silent_tests).returncode
                is_supported = (val == EXIT_SUCCESS)
            except Exception:
                return None
        except Exception:
            pass
        try:
            os.remove(fname)
        except Exception:
            pass
        
        if supports_restrict:
            for e in self.extensions:
                e.define_macros += [("SUPPORTS_RESTRICT", "1")]



setup(
    name  = "recometrics",
    packages = ["recometrics"],
    version = '0.1.6-11',
    cmdclass = {'build_ext': build_ext_subclass},
    author = 'David Cortes',
    url = 'https://github.com/david-cortes/recometrics',
    install_requires = ['numpy', 'scipy', 'cython', 'pandas'],
    description = 'Library-agnostic evaluation framework for implicit-feedback recommender systems',
    ext_modules = [
        Extension("recometrics.cpp_funs",
            language="c++",
            sources=["recometrics/wrapper.pyx", "src/recometrics_instantiated.cpp"],
            include_dirs=[np.get_include(), "src"],
            define_macros=[
                ("_FOR_PYTHON", None),
                ("NDEBUG", None),
                ("CYTHON_EXTERN_C", 'extern "C"'),
            ]
            )
        ]
)

if not found_omp:
    omp_msg  = "\n\n\nCould not detect OpenMP. Package will be built without multi-threading capabilities. "
    omp_msg += " To enable multi-threading, first install OpenMP"
    if (sys.platform[:3] == "dar"):
        omp_msg += " - for macOS: 'brew install libomp'\n"
    else:
        omp_msg += " modules for your compiler. "
    
    omp_msg += "Then reinstall this package from scratch: 'pip install --upgrade --no-deps --force-reinstall recometrics'.\n"
    warnings.warn(omp_msg)



#cython: language_level=3
import numpy as np
cimport numpy as np
from libcpp cimport bool as c_bool
from libcpp.vector cimport vector
from libc.string cimport memcpy
from libc.stdint cimport int32_t, uint64_t
from scipy.sparse import csr_array
import ctypes
from scipy.linalg.cython_blas cimport ddot, sdot

### TODO: here cython fails to compile the functions using fused types or templates,
### and depending on how it's done (fused/templates), might produce something compilable
### with gcc/clang but not with msvc. Find out why it fails and avoid this duplication
### of code. Perhaps it has to do with this issue:
### https://github.com/cython/cython/issues/3968

ctypedef fused generic_t:
    int32_t
    double
    float

ctypedef fused real_t:
    double
    float

ctypedef double (*ddot__)(const int*, const double*, const int*, const double*, const int*) noexcept nogil
ctypedef float (*sdot__)(const int*, const float*, const int*, const float*, const int*) noexcept nogil

cdef public double ddot_(const int* n, const double* x, const int* incx, const double* y, const int* incy) noexcept nogil:
    return (<ddot__>ddot)(n, x, incx, y, incy)

cdef public float sdot_(const int* n, const float* x, const int* incx, const float* y, const int* incy) noexcept nogil:
    return (<sdot__>sdot)(n, x, incx, y, incy)

cdef extern from "recometrics_signatures.hpp":
    c_bool get_has_openmp() except + nogil

    void calc_metrics_double(
        const double *A, const size_t lda, const double *B, const size_t ldb,
        const int32_t m, const int32_t n, const int32_t k,
        const int32_t *Xtrain_csr_p, const int32_t *Xtrain_csr_i,
        const int32_t *Xtest_csr_p, const int32_t *Xtest_csr_i, const double *Xtest_csr,
        const int32_t k_metrics,
        const c_bool cumulative,
        const c_bool break_ties_with_noise,
        double *p_at_k,
        double *tp_at_k,
        double *r_at_k,
        double *ap_at_k,
        double *tap_at_k,
        double *ndcg_at_k,
        double *hit_at_k,
        double *rr_at_k,
        double *roc_auc,
        double *pr_auc,
        const c_bool consider_cold_start,
        int32_t min_items_pool,
        int32_t min_pos_test,
        int32_t nthreads,
        uint64_t seed
    ) except + nogil

    void split_data_selected_users_double(
        const int32_t *X_csr_p,
        const int32_t *X_csr_i,
        const double *X_csr,
        const int32_t m, const int32_t n,
        vector[int32_t] &Xtrain_csr_p,
        vector[int32_t] &Xtrain_csr_i,
        vector[double] &Xtrain_csr,
        vector[int32_t] &Xtest_csr_p,
        vector[int32_t] &Xtest_csr_i,
        vector[double] &Xtest_csr,
        const double test_fraction,
        uint64_t seed
    ) except + nogil

    void split_data_separate_users_double(
        const int32_t *X_csr_p,
        const int32_t *X_csr_i,
        const double *X_csr,
        int32_t m, int32_t n,
        vector[int32_t] &users_test,
        vector[int32_t] &Xrem_csr_p,
        vector[int32_t] &Xrem_csr_i,
        vector[double] &Xrem_csr,
        vector[int32_t] &Xtrain_csr_p,
        vector[int32_t] &Xtrain_csr_i,
        vector[double] &Xtrain_csr,
        vector[int32_t] &Xtest_csr_p,
        vector[int32_t] &Xtest_csr_i,
        vector[double] &Xtest_csr,
        const int32_t n_users_test,
        const double test_fraction,
        const c_bool consider_cold_start,
        const int32_t min_items_pool,
        const int32_t min_pos_test,
        uint64_t seed
    ) except + nogil

    void split_data_joined_users_double(
        const int32_t *X_csr_p,
        const int32_t *X_csr_i,
        const double *X_csr,
        int32_t m, int32_t n,
        vector[int32_t] &users_test,
        vector[int32_t] &Xtrain_csr_p,
        vector[int32_t] &Xtrain_csr_i,
        vector[double] &Xtrain_csr,
        vector[int32_t] &Xtest_csr_p,
        vector[int32_t] &Xtest_csr_i,
        vector[double] &Xtest_csr,
        const int32_t n_users_test,
        const double test_fraction,
        const c_bool consider_cold_start,
        const int32_t min_items_pool,
        const int32_t min_pos_test,
        uint64_t seed
    ) except + nogil

    void calc_metrics_float(
        const float *A, const size_t lda, const float *B, const size_t ldb,
        const int32_t m, const int32_t n, const int32_t k,
        const int32_t *Xtrain_csr_p, const int32_t *Xtrain_csr_i,
        const int32_t *Xtest_csr_p, const int32_t *Xtest_csr_i, const float *Xtest_csr,
        const int32_t k_metrics,
        const c_bool cumulative,
        const c_bool break_ties_with_noise,
        float *p_at_k,
        float *tp_at_k,
        float *r_at_k,
        float *ap_at_k,
        float *tap_at_k,
        float *ndcg_at_k,
        float *hit_at_k,
        float *rr_at_k,
        float *roc_auc,
        float *pr_auc,
        const c_bool consider_cold_start,
        int32_t min_items_pool,
        int32_t min_pos_test,
        int32_t nthreads,
        uint64_t seed
    ) except + nogil

    void split_data_selected_users_float(
        const int32_t *X_csr_p,
        const int32_t *X_csr_i,
        const float *X_csr,
        const int32_t m, const int32_t n,
        vector[int32_t] &Xtrain_csr_p,
        vector[int32_t] &Xtrain_csr_i,
        vector[float] &Xtrain_csr,
        vector[int32_t] &Xtest_csr_p,
        vector[int32_t] &Xtest_csr_i,
        vector[float] &Xtest_csr,
        const double test_fraction,
        uint64_t seed
    ) except + nogil

    void split_data_separate_users_float(
        const int32_t *X_csr_p,
        const int32_t *X_csr_i,
        const float *X_csr,
        int32_t m, int32_t n,
        vector[int32_t] &users_test,
        vector[int32_t] &Xrem_csr_p,
        vector[int32_t] &Xrem_csr_i,
        vector[float] &Xrem_csr,
        vector[int32_t] &Xtrain_csr_p,
        vector[int32_t] &Xtrain_csr_i,
        vector[float] &Xtrain_csr,
        vector[int32_t] &Xtest_csr_p,
        vector[int32_t] &Xtest_csr_i,
        vector[float] &Xtest_csr,
        const int32_t n_users_test,
        const double test_fraction,
        const c_bool consider_cold_start,
        const int32_t min_items_pool,
        const int32_t min_pos_test,
        uint64_t seed
    ) except + nogil

    void split_data_joined_users_float(
        const int32_t *X_csr_p,
        const int32_t *X_csr_i,
        const float *X_csr,
        int32_t m, int32_t n,
        vector[int32_t] &users_test,
        vector[int32_t] &Xtrain_csr_p,
        vector[int32_t] &Xtrain_csr_i,
        vector[float] &Xtrain_csr,
        vector[int32_t] &Xtest_csr_p,
        vector[int32_t] &Xtest_csr_i,
        vector[float] &Xtest_csr,
        const int32_t n_users_test,
        const double test_fraction,
        const c_bool consider_cold_start,
        const int32_t min_items_pool,
        const int32_t min_pos_test,
        uint64_t seed
    ) except + nogil

def _get_has_openmp():
    return get_has_openmp()

cdef double* get_ptr_double(np.ndarray[double, ndim=1] a):
    if a.shape[0]:
        return &a[0]
    else:
        return <double*>NULL

cdef float* get_ptr_float(np.ndarray[float, ndim=1] a):
    if a.shape[0]:
        return &a[0]
    else:
        return <float*>NULL

cdef int32_t* get_ptr_int(np.ndarray[int32_t, ndim=1] a):
    if a.shape[0]:
        return &a[0]
    else:
        return <int32_t*>NULL

cdef tuple cy_calc_metrics_double(
    np.ndarray[double, ndim=2] A, size_t lda,
    np.ndarray[double, ndim=2] B, size_t ldb,
    Xcsr_train, Xcsr_test,
    int32_t k_metrics = 10,
    c_bool precision = True,
    c_bool trunc_precision = False,
    c_bool recall = False,
    c_bool average_precision = True,
    c_bool trunc_average_precision = False,
    c_bool ndcg = True,
    c_bool hit = False,
    c_bool rr = False,
    c_bool roc_auc = False,
    c_bool pr_auc = False,
    c_bool break_ties_with_noise = 1,
    int32_t min_pos_test = 1,
    int32_t min_items_pool = 2,
    c_bool consider_cold_start = 0,
    c_bool cumulative = 0,
    int32_t nthreads = 1,
    uint64_t seed = 1
):
    
    cdef np.ndarray[int32_t, ndim=1] Xtrain_csr_p = Xcsr_train.indptr
    cdef np.ndarray[int32_t, ndim=1] Xtrain_csr_i = Xcsr_train.indices

    cdef np.ndarray[int32_t, ndim=1] Xtest_csr_p = Xcsr_test.indptr
    cdef np.ndarray[int32_t, ndim=1] Xtest_csr_i = Xcsr_test.indices
    cdef np.ndarray[double, ndim=1] Xtest_csr = Xcsr_test.data

    cdef int32_t m = A.shape[0]
    cdef int32_t n = B.shape[0]
    cdef int32_t k = A.shape[1]

    if (A.shape[0] != Xcsr_train.shape[0]):
        raise ValueError("User shapes do not match.")
    if (B.shape[0] != Xcsr_train.shape[1]):
        raise ValueError("Item shapes do not match.")
    if (A.shape[1] != B.shape[1]):
        raise ValueError("Factors have non-matching shapes.")
    if (Xcsr_train.shape[0] != Xcsr_test.shape[0]) or (Xcsr_train.shape[1] != Xcsr_test.shape[1]):
        raise ValueError("Train-test shapes do not match.")

    cdef size_t size_arr = m if not cumulative else m*k_metrics
    cdef np.ndarray[double, ndim=1] p_at_k = np.empty(size_arr if precision else 0, dtype=ctypes.c_double)
    cdef np.ndarray[double, ndim=1] tp_at_k = np.empty(size_arr if trunc_precision else 0, dtype=ctypes.c_double)
    cdef np.ndarray[double, ndim=1] r_at_k = np.empty(size_arr if recall else 0, dtype=ctypes.c_double)
    cdef np.ndarray[double, ndim=1] ap_at_k = np.empty(size_arr if average_precision else 0, dtype=ctypes.c_double)
    cdef np.ndarray[double, ndim=1] tap_at_k = np.empty(size_arr if trunc_average_precision else 0, dtype=ctypes.c_double)
    cdef np.ndarray[double, ndim=1] ndcg_at_k = np.empty(size_arr if ndcg else 0, dtype=ctypes.c_double)
    cdef np.ndarray[double, ndim=1] hit_at_k = np.empty(size_arr if hit else 0, dtype=ctypes.c_double)
    cdef np.ndarray[double, ndim=1] rr_at_k = np.empty(size_arr if rr else 0, dtype=ctypes.c_double)
    cdef np.ndarray[double, ndim=1] roc_auc_ = np.empty(m if roc_auc else 0, dtype=ctypes.c_double)
    cdef np.ndarray[double, ndim=1] pr_auc_ = np.empty(m if pr_auc else 0, dtype=ctypes.c_double)

    calc_metrics_double(
        &A[0,0], lda, &B[0,0], ldb, m, n, k,
        get_ptr_int(Xtrain_csr_p), get_ptr_int(Xtrain_csr_i),
        get_ptr_int(Xtest_csr_p), get_ptr_int(Xtest_csr_i), &Xtest_csr[0],
        k_metrics,
        cumulative,
        break_ties_with_noise,
        get_ptr_double(p_at_k),
        get_ptr_double(tp_at_k),
        get_ptr_double(r_at_k),
        get_ptr_double(ap_at_k),
        get_ptr_double(tap_at_k),
        get_ptr_double(ndcg_at_k),
        get_ptr_double(hit_at_k),
        get_ptr_double(rr_at_k),
        get_ptr_double(roc_auc_),
        get_ptr_double(pr_auc_),
        consider_cold_start,
        min_items_pool,
        min_pos_test,
        nthreads,
        seed
    )

    cdef tuple size_shape = (m,k_metrics)
    cdef tuple size_empty = (<int>0,<int>0)

    if not cumulative:
        return p_at_k, tp_at_k, r_at_k, ap_at_k, tap_at_k, ndcg_at_k, hit_at_k, rr_at_k, roc_auc_, pr_auc_
    else:
        return (
            p_at_k.reshape(size_shape if precision else size_empty),
            tp_at_k.reshape(size_shape if trunc_precision else size_empty),
            r_at_k.reshape(size_shape if recall else size_empty),
            ap_at_k.reshape(size_shape if average_precision else size_empty),
            tap_at_k.reshape(size_shape if trunc_average_precision else size_empty),
            ndcg_at_k.reshape(size_shape if ndcg else size_empty),
            hit_at_k.reshape(size_shape if hit else size_empty),
            rr_at_k.reshape(size_shape if rr else size_empty),
            roc_auc_,
            pr_auc_
        )

cdef tuple cy_calc_metrics_float(
    np.ndarray[float, ndim=2] A, size_t lda,
    np.ndarray[float, ndim=2] B, size_t ldb,
    Xcsr_train, Xcsr_test,
    int32_t k_metrics = 10,
    c_bool precision = True,
    c_bool trunc_precision = False,
    c_bool recall = False,
    c_bool average_precision = True,
    c_bool trunc_average_precision = False,
    c_bool ndcg = True,
    c_bool hit = False,
    c_bool rr = False,
    c_bool roc_auc = False,
    c_bool pr_auc = False,
    c_bool break_ties_with_noise = 1,
    int32_t min_pos_test = 1,
    int32_t min_items_pool = 2,
    c_bool consider_cold_start = 0,
    c_bool cumulative = 0,
    int32_t nthreads = 1,
    uint64_t seed = 1
):
    
    cdef np.ndarray[int32_t, ndim=1] Xtrain_csr_p = Xcsr_train.indptr
    cdef np.ndarray[int32_t, ndim=1] Xtrain_csr_i = Xcsr_train.indices

    cdef np.ndarray[int32_t, ndim=1] Xtest_csr_p = Xcsr_test.indptr
    cdef np.ndarray[int32_t, ndim=1] Xtest_csr_i = Xcsr_test.indices
    cdef np.ndarray[float, ndim=1] Xtest_csr = Xcsr_test.data

    cdef int32_t m = A.shape[0]
    cdef int32_t n = B.shape[0]
    cdef int32_t k = A.shape[1]

    if (A.shape[0] != Xcsr_train.shape[0]):
        raise ValueError("User shapes do not match.")
    if (B.shape[0] != Xcsr_train.shape[1]):
        raise ValueError("Item shapes do not match.")
    if (A.shape[1] != B.shape[1]):
        raise ValueError("Factors have non-matching shapes.")
    if (Xcsr_train.shape[0] != Xcsr_test.shape[0]) or (Xcsr_train.shape[1] != Xcsr_test.shape[1]):
        raise ValueError("Train-test shapes do not match.")

    cdef size_t size_arr = m if not cumulative else m*k_metrics
    cdef np.ndarray[float, ndim=1] p_at_k = np.empty(size_arr if precision else 0, dtype=ctypes.c_float)
    cdef np.ndarray[float, ndim=1] tp_at_k = np.empty(size_arr if trunc_precision else 0, dtype=ctypes.c_float)
    cdef np.ndarray[float, ndim=1] r_at_k = np.empty(size_arr if recall else 0, dtype=ctypes.c_float)
    cdef np.ndarray[float, ndim=1] ap_at_k = np.empty(size_arr if average_precision else 0, dtype=ctypes.c_float)
    cdef np.ndarray[float, ndim=1] tap_at_k = np.empty(size_arr if trunc_average_precision else 0, dtype=ctypes.c_float)
    cdef np.ndarray[float, ndim=1] ndcg_at_k = np.empty(size_arr if ndcg else 0, dtype=ctypes.c_float)
    cdef np.ndarray[float, ndim=1] hit_at_k = np.empty(size_arr if hit else 0, dtype=ctypes.c_float)
    cdef np.ndarray[float, ndim=1] rr_at_k = np.empty(size_arr if rr else 0, dtype=ctypes.c_float)
    cdef np.ndarray[float, ndim=1] roc_auc_ = np.empty(m if roc_auc else 0, dtype=ctypes.c_float)
    cdef np.ndarray[float, ndim=1] pr_auc_ = np.empty(m if pr_auc else 0, dtype=ctypes.c_float)

    calc_metrics_float(
        &A[0,0], lda, &B[0,0], ldb, m, n, k,
        get_ptr_int(Xtrain_csr_p), get_ptr_int(Xtrain_csr_i),
        get_ptr_int(Xtest_csr_p), get_ptr_int(Xtest_csr_i), &Xtest_csr[0],
        k_metrics,
        cumulative,
        break_ties_with_noise,
        get_ptr_float(p_at_k),
        get_ptr_float(tp_at_k),
        get_ptr_float(r_at_k),
        get_ptr_float(ap_at_k),
        get_ptr_float(tap_at_k),
        get_ptr_float(ndcg_at_k),
        get_ptr_float(hit_at_k),
        get_ptr_float(rr_at_k),
        get_ptr_float(roc_auc_),
        get_ptr_float(pr_auc_),
        consider_cold_start,
        min_items_pool,
        min_pos_test,
        nthreads,
        seed
    )

    
    cdef tuple size_shape = (m,k_metrics)
    cdef tuple size_empty = (<int>0,<int>0)

    if not cumulative:
        return p_at_k, tp_at_k, r_at_k, ap_at_k, tap_at_k, ndcg_at_k, hit_at_k, rr_at_k, roc_auc_, pr_auc_
    else:
        return (
            p_at_k.reshape(size_shape if precision else size_empty),
            tp_at_k.reshape(size_shape if trunc_precision else size_empty),
            r_at_k.reshape(size_shape if recall else size_empty),
            ap_at_k.reshape(size_shape if average_precision else size_empty),
            tap_at_k.reshape(size_shape if trunc_average_precision else size_empty),
            ndcg_at_k.reshape(size_shape if ndcg else size_empty),
            hit_at_k.reshape(size_shape if hit else size_empty),
            rr_at_k.reshape(size_shape if rr else size_empty),
            roc_auc_,
            pr_auc_
        )

def calc_reco_metrics(
    A, size_t lda,
    B, size_t ldb,
    Xcsr_train, Xcsr_test,
    int32_t k_metrics = 10,
    c_bool precision = True,
    c_bool trunc_precision = False,
    c_bool recall = False,
    c_bool average_precision = True,
    c_bool trunc_average_precision = False,
    c_bool ndcg = True,
    c_bool hit = False,
    c_bool rr = False,
    c_bool roc_auc = False,
    c_bool pr_auc = False,
    c_bool break_ties_with_noise = 1,
    int32_t min_pos_test = 1,
    int32_t min_items_pool = 2,
    c_bool consider_cold_start = 0,
    c_bool cumulative = 0,
    int32_t nthreads = 1,
    uint64_t seed = 1
):
    if A.dtype == ctypes.c_float:
        return cy_calc_metrics_float(
            A, lda,
            B, ldb,
            Xcsr_train, Xcsr_test,
            k_metrics,
            precision,
            trunc_precision,
            recall,
            average_precision,
            trunc_average_precision,
            ndcg,
            hit,
            rr,
            roc_auc,
            pr_auc,
            break_ties_with_noise,
            min_pos_test,
            min_items_pool,
            consider_cold_start,
            cumulative,
            nthreads,
            seed
        )
    else:
        return cy_calc_metrics_double(
            A, lda,
            B, ldb,
            Xcsr_train, Xcsr_test,
            k_metrics,
            precision,
            trunc_precision,
            recall,
            average_precision,
            trunc_average_precision,
            ndcg,
            hit,
            rr,
            roc_auc,
            pr_auc,
            break_ties_with_noise,
            min_pos_test,
            min_items_pool,
            consider_cold_start,
            cumulative,
            nthreads,
            seed
        )

cdef np.ndarray[generic_t, ndim=1] cppvec_to_numpy(
    vector[generic_t] inp
):
    cdef np.ndarray[generic_t, ndim=1] out = np.empty(inp.size(),
        dtype=np.int32 if (generic_t is int32_t) else (ctypes.c_double if generic_t is double else (ctypes.c_float))
    )
    if not inp.size():
        return out
    memcpy(&out[0], inp.data(), inp.size() * <size_t>sizeof(generic_t))
    inp.clear()
    return out

cdef object cpp_csr_to_scipy(
    vector[int32_t] Xcsr_p,
    vector[int32_t] Xcsr_i,
    vector[generic_t] Xcsr,
    int32_t n
):
    cdef np.ndarray[int32_t, ndim=1] Xcsr_p_ = cppvec_to_numpy[int32_t](Xcsr_p)
    cdef np.ndarray[int32_t, ndim=1] Xcsr_i_ = cppvec_to_numpy[int32_t](Xcsr_i)
    cdef np.ndarray[generic_t, ndim=1] Xcsr_ = cppvec_to_numpy[generic_t](Xcsr)
    cdef int m = Xcsr_p_.shape[0] - 1
    return csr_array((Xcsr_, Xcsr_i_, Xcsr_p_), shape=(m, n))

cdef tuple split_csr_selected_users_double(
    X_csr_scipy,
    const double test_fraction,
    uint64_t seed = 1
):
    cdef np.ndarray[int32_t, ndim=1] X_csr_p = X_csr_scipy.indptr
    cdef np.ndarray[int32_t, ndim=1] X_csr_i = X_csr_scipy.indices
    cdef np.ndarray[double, ndim=1] X_csr = X_csr_scipy.data
    cdef int32_t m = X_csr_scipy.shape[0]
    cdef int32_t n = X_csr_scipy.shape[1]

    cdef vector[int32_t] Xtrain_csr_p
    cdef vector[int32_t] Xtrain_csr_i
    cdef vector[double] Xtrain_csr
    cdef vector[int32_t] Xtest_csr_p
    cdef vector[int32_t] Xtest_csr_i
    cdef vector[double] Xtest_csr

    cdef const double* tempp = &X_csr[0] if X_csr.shape[0] else NULL
    
    split_data_selected_users_double(
        get_ptr_int(X_csr_p),
        get_ptr_int(X_csr_i),
        tempp,
        m, n,
        Xtrain_csr_p,
        Xtrain_csr_i,
        Xtrain_csr,
        Xtest_csr_p,
        Xtest_csr_i,
        Xtest_csr,
        test_fraction,
        seed
    )

    Xtrain_out = cpp_csr_to_scipy[double](Xtrain_csr_p, Xtrain_csr_i, Xtrain_csr, n)
    Xtest_out = cpp_csr_to_scipy[double](Xtest_csr_p, Xtest_csr_i, Xtest_csr, n)
    return Xtrain_out, Xtest_out

cdef tuple split_csr_selected_users_float(
    X_csr_scipy,
    const double test_fraction,
    uint64_t seed = 1
):
    cdef np.ndarray[int32_t, ndim=1] X_csr_p = X_csr_scipy.indptr
    cdef np.ndarray[int32_t, ndim=1] X_csr_i = X_csr_scipy.indices
    cdef np.ndarray[float, ndim=1] X_csr = X_csr_scipy.data
    cdef int32_t m = X_csr_scipy.shape[0]
    cdef int32_t n = X_csr_scipy.shape[1]

    cdef vector[int32_t] Xtrain_csr_p
    cdef vector[int32_t] Xtrain_csr_i
    cdef vector[float] Xtrain_csr
    cdef vector[int32_t] Xtest_csr_p
    cdef vector[int32_t] Xtest_csr_i
    cdef vector[float] Xtest_csr

    cdef const float* tempp = &X_csr[0]
    
    split_data_selected_users_float(
        get_ptr_int(X_csr_p),
        get_ptr_int(X_csr_i),
        tempp,
        m, n,
        Xtrain_csr_p,
        Xtrain_csr_i,
        Xtrain_csr,
        Xtest_csr_p,
        Xtest_csr_i,
        Xtest_csr,
        test_fraction,
        seed
    )

    Xtrain_out = cpp_csr_to_scipy[float](Xtrain_csr_p, Xtrain_csr_i, Xtrain_csr, n)
    Xtest_out = cpp_csr_to_scipy[float](Xtest_csr_p, Xtest_csr_i, Xtest_csr, n)
    return Xtrain_out, Xtest_out

def split_csr_selected_users(
    X_csr_scipy,
    const double test_fraction = 0.2,
    uint64_t seed = 1
):
    if X_csr_scipy.dtype == ctypes.c_float:
        return split_csr_selected_users_float(
            X_csr_scipy,
            test_fraction,
            seed
        )
    else:
        return split_csr_selected_users_double(
            X_csr_scipy,
            test_fraction,
            seed
        )

cdef tuple split_csr_separated_users_double(
    X_csr_scipy,
    const int32_t n_users_test,
    const double test_fraction,
    const c_bool consider_cold_start,
    const int32_t min_items_pool,
    const int32_t min_pos_test,
    const c_bool separated,
    uint64_t seed
):
    cdef np.ndarray[int32_t, ndim=1] X_csr_p = X_csr_scipy.indptr
    cdef np.ndarray[int32_t, ndim=1] X_csr_i = X_csr_scipy.indices
    cdef np.ndarray[double, ndim=1] X_csr = X_csr_scipy.data
    cdef int32_t m = X_csr_scipy.shape[0]
    cdef int32_t n = X_csr_scipy.shape[1]

    cdef vector[int32_t] Xtrain_csr_p
    cdef vector[int32_t] Xtrain_csr_i
    cdef vector[double] Xtrain_csr
    cdef vector[int32_t] Xtest_csr_p
    cdef vector[int32_t] Xtest_csr_i
    cdef vector[double] Xtest_csr
    cdef vector[int32_t] Xrem_csr_p
    cdef vector[int32_t] Xrem_csr_i
    cdef vector[double] Xrem_csr

    cdef vector[int32_t] users_test
    cdef np.ndarray[int32_t, ndim=1] users_test_

    if separated:
        split_data_separate_users_double(
            get_ptr_int(X_csr_p),
            get_ptr_int(X_csr_i),
            &X_csr[0] if X_csr.shape[0] else NULL,
            m, n,
            users_test,
            Xrem_csr_p,
            Xrem_csr_i,
            Xrem_csr,
            Xtrain_csr_p,
            Xtrain_csr_i,
            Xtrain_csr,
            Xtest_csr_p,
            Xtest_csr_i,
            Xtest_csr,
            n_users_test,
            test_fraction,
            consider_cold_start,
            min_items_pool,
            min_pos_test,
            seed
        )

        Xtrain_out = cpp_csr_to_scipy[double](Xtrain_csr_p, Xtrain_csr_i, Xtrain_csr, n)
        Xtest_out = cpp_csr_to_scipy[double](Xtest_csr_p, Xtest_csr_i, Xtest_csr, n)
        Xrem_out = cpp_csr_to_scipy[double](Xrem_csr_p, Xrem_csr_i, Xrem_csr, n)
        users_test_ = cppvec_to_numpy[int32_t](users_test)
        return Xrem_out, Xtrain_out, Xtest_out, users_test_

    else:
        split_data_joined_users_double(
            get_ptr_int(X_csr_p),
            get_ptr_int(X_csr_i),
            &X_csr[0] if X_csr.shape[0] else NULL,
            m, n,
            users_test,
            Xtrain_csr_p,
            Xtrain_csr_i,
            Xtrain_csr,
            Xtest_csr_p,
            Xtest_csr_i,
            Xtest_csr,
            n_users_test,
            test_fraction,
            consider_cold_start,
            min_items_pool,
            min_pos_test,
            seed
        )

        Xtrain_out = cpp_csr_to_scipy[double](Xtrain_csr_p, Xtrain_csr_i, Xtrain_csr, n)
        Xtest_out = cpp_csr_to_scipy[double](Xtest_csr_p, Xtest_csr_i, Xtest_csr, n)
        users_test_ = cppvec_to_numpy[int32_t](users_test)
        return Xtrain_out, Xtest_out, users_test_

cdef tuple split_csr_separated_users_float(
    X_csr_scipy,
    const int32_t n_users_test,
    const double test_fraction,
    const c_bool consider_cold_start,
    const int32_t min_items_pool,
    const int32_t min_pos_test,
    const c_bool separated,
    uint64_t seed
):
    cdef np.ndarray[int32_t, ndim=1] X_csr_p = X_csr_scipy.indptr
    cdef np.ndarray[int32_t, ndim=1] X_csr_i = X_csr_scipy.indices
    cdef np.ndarray[float, ndim=1] X_csr = X_csr_scipy.data
    cdef int32_t m = X_csr_scipy.shape[0]
    cdef int32_t n = X_csr_scipy.shape[1]

    cdef vector[int32_t] Xtrain_csr_p
    cdef vector[int32_t] Xtrain_csr_i
    cdef vector[float] Xtrain_csr
    cdef vector[int32_t] Xtest_csr_p
    cdef vector[int32_t] Xtest_csr_i
    cdef vector[float] Xtest_csr
    cdef vector[int32_t] Xrem_csr_p
    cdef vector[int32_t] Xrem_csr_i
    cdef vector[float] Xrem_csr

    cdef vector[int32_t] users_test
    cdef np.ndarray[int32_t, ndim=1] users_test_

    if separated:
        split_data_separate_users_float(
            get_ptr_int(X_csr_p),
            get_ptr_int(X_csr_i),
            &X_csr[0] if X_csr.shape[0] else NULL,
            m, n,
            users_test,
            Xrem_csr_p,
            Xrem_csr_i,
            Xrem_csr,
            Xtrain_csr_p,
            Xtrain_csr_i,
            Xtrain_csr,
            Xtest_csr_p,
            Xtest_csr_i,
            Xtest_csr,
            n_users_test,
            test_fraction,
            consider_cold_start,
            min_items_pool,
            min_pos_test,
            seed
        )

        Xtrain_out = cpp_csr_to_scipy[float](Xtrain_csr_p, Xtrain_csr_i, Xtrain_csr, n)
        Xtest_out = cpp_csr_to_scipy[float](Xtest_csr_p, Xtest_csr_i, Xtest_csr, n)
        Xrem_out = cpp_csr_to_scipy[float](Xrem_csr_p, Xrem_csr_i, Xrem_csr, n)
        users_test_ = cppvec_to_numpy[int32_t](users_test)
        return Xrem_out, Xtrain_out, Xtest_out, users_test_

    else:
        split_data_joined_users_float(
            get_ptr_int(X_csr_p),
            get_ptr_int(X_csr_i),
            &X_csr[0] if X_csr.shape[0] else NULL,
            m, n,
            users_test,
            Xtrain_csr_p,
            Xtrain_csr_i,
            Xtrain_csr,
            Xtest_csr_p,
            Xtest_csr_i,
            Xtest_csr,
            n_users_test,
            test_fraction,
            consider_cold_start,
            min_items_pool,
            min_pos_test,
            seed
        )

        Xtrain_out = cpp_csr_to_scipy[float](Xtrain_csr_p, Xtrain_csr_i, Xtrain_csr, n)
        Xtest_out = cpp_csr_to_scipy[float](Xtest_csr_p, Xtest_csr_i, Xtest_csr, n)
        users_test_ = cppvec_to_numpy[int32_t](users_test)
        return Xtrain_out, Xtest_out, users_test_

def split_csr_separated_users(
    X_csr_scipy,
    const int32_t n_users_test,
    const double test_fraction = 0.2,
    const c_bool consider_cold_start = 0,
    const int32_t min_items_pool = 10,
    const int32_t min_pos_test = 1,
    const c_bool separated = True,
    uint64_t seed = 1
):
    if X_csr_scipy.dtype == ctypes.c_float:
        return split_csr_separated_users_float(
            X_csr_scipy,
            n_users_test,
            test_fraction,
            consider_cold_start,
            min_items_pool,
            min_pos_test,
            separated,
            seed
        )
    else:
        return split_csr_separated_users_double(
            X_csr_scipy,
            n_users_test,
            test_fraction,
            consider_cold_start,
            min_items_pool,
            min_pos_test,
            separated,
            seed
        )
